@@ -1,31 +1,31 @@
-import random
 import re
-import logging
 import json
-import time
-from pytz import timezone
 import uuid
 import base64
 import codecs
+import random
+import logging
+import datetime
 import xmltodict
 import collections
-import botocore.config
-import six
-import datetime
 import dateutil.parser
-from six.moves.urllib import parse as urlparse
+import six
+import botocore.config
+from pytz import timezone
 from botocore.client import ClientError
 from requests.models import Response, Request
+from six.moves.urllib import parse as urlparse
 from localstack import config, constants
 from localstack.config import HOSTNAME, HOSTNAME_EXTERNAL
 from localstack.utils.aws import aws_stack
+from localstack.services.s3 import multipart_content
 from localstack.utils.common import (
-    short_uid, timestamp_millis, to_str, to_bytes, clone, md5, get_service_protocol
+    short_uid, timestamp_millis, to_str, to_bytes, clone, md5, get_service_protocol, now_utc
 )
 from localstack.utils.analytics import event_publisher
-from localstack.utils.aws.aws_responses import requests_response
+from localstack.utils.http_utils import uses_chunked_encoding
 from localstack.utils.persistence import PersistingProxyListener
-from localstack.services.s3 import multipart_content
+from localstack.utils.aws.aws_responses import requests_response
 
 CONTENT_SHA256_HEADER = 'x-amz-content-sha256'
 STREAMING_HMAC_PAYLOAD = 'STREAMING-AWS4-HMAC-SHA256-PAYLOAD'
@@ -80,6 +80,10 @@ ALLOWED_HEADER_OVERRIDES = {
     'response-content-encoding': 'Content-Encoding',
 }
 
+# From botocore's auth.py:
+# https://github.com/boto/botocore/blob/30206ab9e9081c80fa68e8b2cb56296b09be6337/botocore/auth.py#L47
+POLICY_EXPIRATION_FORMAT = '%Y-%m-%dT%H:%M:%SZ'
+
 
 def event_type_matches(events, action, api_method):
     """ check whether any of the event types in `events` matches the
@@ -288,6 +292,12 @@ def delete_cors(bucket_name):
     return response
 
 
+def convert_origins_into_list(allowed_origins):
+    if isinstance(allowed_origins, list):
+        return allowed_origins
+    return [allowed_origins]
+
+
 def append_cors_headers(bucket_name, request_method, request_headers, response):
     bucket_name = normalize_bucket_name(bucket_name)
 
@@ -304,6 +314,9 @@ def append_cors_headers(bucket_name, request_method, request_headers, response):
         allowed_methods = rule.get('AllowedMethod', [])
         if request_method in allowed_methods:
             allowed_origins = rule.get('AllowedOrigin', [])
+            # when only one origin is being set in cors then the allowed_origins is being
+            # reflected as a string here,so making it a list and then proceeding.
+            allowed_origins = convert_origins_into_list(allowed_origins)
             for allowed in allowed_origins:
                 if origin in allowed or re.match(allowed.replace('*', '.*'), origin):
                     response.headers['Access-Control-Allow-Origin'] = origin
@@ -346,7 +359,7 @@ def get_object_expiry(path):
 
 
 def is_url_already_expired(expiry_timestamp):
-    if int(expiry_timestamp) < int(time.time()):
+    if int(expiry_timestamp) < int(now_utc()):
         return True
     return False
 
@@ -357,7 +370,8 @@ def add_reponse_metadata_headers(response):
     if response.headers.get('cache-control') is None:
         response.headers['cache-control'] = 'no-cache'
     if response.headers.get('content-encoding') is None:
-        response.headers['content-encoding'] = 'identity'
+        if not uses_chunked_encoding(response):
+            response.headers['content-encoding'] = 'identity'
 
 
 def append_last_modified_headers(response, content=None):
@@ -395,7 +409,7 @@ def append_list_objects_marker(method, path, data, response):
             query_map = urlparse.parse_qs(parsed.query)
             insert = '<Marker>%s</Marker>' % query_map.get('marker')[0]
             response._content = content.replace('</ListBucketResult>', '%s</ListBucketResult>' % insert)
-            response.headers['Content-Length'] = str(len(response._content))
+            response.headers.pop('Content-Length', None)
 
 
 def append_metadata_headers(method, query_map, headers):
@@ -479,8 +493,18 @@ def fix_metadata_key_underscores(request_headers={}, response=None):
 def fix_creation_date(method, path, response):
     if method != 'GET' or path != '/':
         return
+    response._content = re.sub(r'(\.[0-9]+)(\+00:00)?</CreationDate>',
+        r'\1Z</CreationDate>', to_str(response._content))
 
-    response._content = re.sub(r'([0-9])</CreationDate>', r'\1Z</CreationDate>', to_str(response._content))
+
+def convert_to_chunked_encoding(method, path, response):
+    if method != 'GET' or path != '/':
+        return
+    if response.headers.get('Transfer-Encoding', '').lower() == 'chunked':
+        return
+    response.headers['Transfer-Encoding'] = 'chunked'
+    response.headers.pop('Content-Encoding', None)
+    response.headers.pop('Content-Length', None)
 
 
 def fix_etag_for_multipart(data, headers, response):
@@ -509,6 +533,7 @@ def remove_xml_preamble(response):
 # HELPER METHODS
 #   for lifecycle/replication/encryption/...
 # --------------
+
 def get_lifecycle(bucket_name):
     bucket_name = normalize_bucket_name(bucket_name)
     exists, code, body = is_bucket_available(bucket_name)
@@ -641,6 +666,7 @@ def set_object_lock(bucket_name, lock_config):
 # -------------
 # UTIL METHODS
 # -------------
+
 def strip_chunk_signatures(data):
     # For clients that use streaming v4 authentication, the request contains chunk signatures
     # in the HTTP body (see example below) which we need to strip as moto cannot handle them
@@ -651,7 +677,7 @@ def strip_chunk_signatures(data):
     data_new = ''
     if data is not None:
         data_new = re.sub(b'(^|\r\n)[0-9a-fA-F]+;chunk-signature=[0-9a-f]{64}(\r\n)(\r\n$)?', b'',
-        data, flags=re.MULTILINE | re.DOTALL)
+            to_bytes(data), flags=re.MULTILINE | re.DOTALL)
 
     return data_new
 
@@ -689,13 +715,13 @@ def bucket_exists(bucket_name):
 
 def check_content_md5(data, headers):
     actual = md5(strip_chunk_signatures(data))
-    expected = headers['Content-MD5']
     try:
-        expected = to_str(codecs.encode(base64.b64decode(expected), 'hex'))
+        expected = to_str(codecs.encode(base64.b64decode(headers['Content-MD5']), 'hex'))
     except Exception:
-        expected = '__invalid__'
+        return error_response('The Content-MD5 you specified is not valid.', 'InvalidDigest', status_code=400)
     if actual != expected:
-        return error_response('The Content-MD5 you specified was invalid', 'InvalidDigest', status_code=400)
+        return error_response('The Content-MD5 you specified did not match what we received.',
+            'BadDigest', status_code=400)
 
 
 def error_response(message, code, status_code=400):
@@ -865,6 +891,10 @@ def remove_bucket_notification(bucket):
     S3_NOTIFICATIONS.pop(bucket, None)
 
 
+def not_none_or(value, alternative):
+    return value if value is not None else alternative
+
+
 class ProxyListenerS3(PersistingProxyListener):
     def api_name(self):
         return 's3'
@@ -941,22 +971,26 @@ class ProxyListenerS3(PersistingProxyListener):
             return error_response('The specified bucket is not valid.', 'InvalidBucketName', status_code=400)
 
         # TODO: For some reason, moto doesn't allow us to put a location constraint on us-east-1
-        to_find = to_bytes('<LocationConstraint>us-east-1</LocationConstraint>')
-        if data and data.startswith(to_bytes('<')) and to_find in data:
-            modified_data = data.replace(to_find, to_bytes(''))
+        to_find1 = to_bytes('<LocationConstraint>us-east-1</LocationConstraint>')
+        to_find2 = to_bytes('<CreateBucketConfiguration')
+        if data and data.startswith(to_bytes('<')) and to_find1 in data and to_find2 in data:
+            # Note: with the latest version, <CreateBucketConfiguration> must either
+            # contain a valid <LocationConstraint>, or not be present at all in the body.
+            modified_data = b''
 
         # If this request contains streaming v4 authentication signatures, strip them from the message
         # Related isse: https://github.com/localstack/localstack/issues/98
         # TODO we should evaluate whether to replace moto s3 with scality/S3:
         # https://github.com/scality/S3/issues/237
-        if headers.get(CONTENT_SHA256_HEADER) == STREAMING_HMAC_PAYLOAD:
-            modified_data = strip_chunk_signatures(modified_data or data)
-            headers['content-length'] = headers.get('x-amz-decoded-content-length')
+        is_streaming_payload = headers.get(CONTENT_SHA256_HEADER) == STREAMING_HMAC_PAYLOAD
+        if is_streaming_payload:
+            modified_data = strip_chunk_signatures(not_none_or(modified_data, data))
+            headers['Content-Length'] = headers.get('x-amz-decoded-content-length')
 
         # POST requests to S3 may include a "${filename}" placeholder in the
         # key, which should be replaced with an actual file name before storing.
         if method == 'POST':
-            original_data = modified_data or data
+            original_data = not_none_or(modified_data, data)
             expanded_data = multipart_content.expand_multipart_filename(original_data, headers)
             if expanded_data is not original_data:
                 modified_data = expanded_data
@@ -988,6 +1022,18 @@ class ProxyListenerS3(PersistingProxyListener):
             if is_url_already_expired(query_map.get('Expires')[0]):
                 return token_expired_error(path, headers.get('x-amz-request-id'), 400)
 
+        # If multipart POST with policy in the params, return error if the policy has expired
+        if method == 'POST':
+            policy_key, policy_value = multipart_content.find_multipart_key_value(data, headers, 'policy')
+            if policy_key and policy_value:
+                policy = json.loads(base64.b64decode(policy_value).decode('utf-8'))
+                expiration_string = policy.get('expiration', None)  # Example: 2020-06-05T13:37:12Z
+                if expiration_string:
+                    expiration_datetime = datetime.datetime.strptime(expiration_string, POLICY_EXPIRATION_FORMAT)
+                    expiration_timestamp = expiration_datetime.timestamp()
+                    if is_url_already_expired(expiration_timestamp):
+                        return token_expired_error(path, headers.get('x-amz-request-id'), 400)
+
         if query == 'cors' or 'cors' in query_map:
             if method == 'GET':
                 return get_cors(bucket)
@@ -1021,7 +1067,10 @@ class ProxyListenerS3(PersistingProxyListener):
                 return set_object_lock(bucket, data)
 
         if modified_data is not None or headers_changed:
-            return Request(data=modified_data or data, headers=headers, method=method)
+            data_to_return = not_none_or(modified_data, data)
+            if modified_data is not None:
+                headers['Content-Length'] = str(len(data_to_return or ''))
+            return Request(data=data_to_return, headers=headers, method=method)
         return True
 
     def get_forward_url(self, method, path, data, headers):
@@ -1040,7 +1089,6 @@ class ProxyListenerS3(PersistingProxyListener):
     def return_response(self, method, path, data, headers, response, request_handler=None):
         path = to_str(path)
         method = to_str(method)
-
         # persist this API call to disk
         super(ProxyListenerS3, self).return_response(method, path, data, headers, response, request_handler)
 
@@ -1072,6 +1120,8 @@ class ProxyListenerS3(PersistingProxyListener):
                 response.headers['Content-Length'] = str(len(response._content))
                 response.headers['Content-Type'] = 'application/xml; charset=utf-8'
                 return response
+        if method == 'GET' and response.status_code == 416:
+            return error_response('The requested range cannot be satisfied.', 'InvalidRange', 416)
 
         parsed = urlparse.urlparse(path)
         bucket_name_in_host = headers['host'].startswith(bucket_name)
@@ -1125,12 +1175,12 @@ class ProxyListenerS3(PersistingProxyListener):
                     error_object = s3_client.get_object(Bucket=bucket_name, Key=error_doc_key)
                     response.status_code = 200
                     response._content = error_object['Body'].read()
-                    response.headers['content-length'] = len(response._content)
+                    response.headers['Content-Length'] = str(len(response._content))
             except ClientError:
                 # Pass on the 404 as usual
                 pass
 
-        if response:
+        if response is not None:
             reset_content_length = False
             # append CORS headers and other annotations/patches to response
             append_cors_headers(bucket_name, request_method=method, request_headers=headers, response=response)
@@ -1208,7 +1258,10 @@ class ProxyListenerS3(PersistingProxyListener):
                 reset_content_length = True
 
             if reset_content_length:
-                response.headers['content-length'] = len(response._content)
+                response.headers['Content-Length'] = str(len(response._content))
+
+            # convert to chunked encoding, for compatibility with certain SDKs (e.g., AWS PHP SDK)
+            convert_to_chunked_encoding(method, path, response)
 
 
 # instantiate listener
