@@ -205,8 +205,8 @@ class JsonObject(object):
         return result
 
     @classmethod
-    def from_json_list(cls, l):
-        return [cls.from_json(j) for j in l]
+    def from_json_list(cls, json_list):
+        return [cls.from_json(j) for j in json_list]
 
     @classmethod
     def as_dict(cls, obj):
@@ -292,16 +292,26 @@ class CaptureOutput(object):
         return threading.currentThread().ident
 
     def stdout(self):
-        return self._stdout.getvalue() if hasattr(self._stdout, 'getvalue') else self._stdout
+        return self._stream_value(self._stdout)
 
     def stderr(self):
-        return self._stderr.getvalue() if hasattr(self._stderr, 'getvalue') else self._stderr
+        return self._stream_value(self._stderr)
+
+    def _stream_value(self, stream):
+        return stream.getvalue() if hasattr(stream, 'getvalue') else stream
 
 
 # ----------------
 # UTILITY METHODS
 # ----------------
 
+def start_thread(method, *args, **kwargs):
+    thread = FuncThread(method, *args, **kwargs)
+    thread.start()
+    TMP_THREADS.append(thread)
+    return thread
+
+
 def synchronized(lock=None):
     """
     Synchronization decorator as described in
@@ -357,6 +367,10 @@ def get_docker_container_names():
     return bootstrap.get_docker_container_names()
 
 
+def path_from_url(url):
+    return '/%s' % str(url).partition('://')[2].partition('/')[2] if '://' in url else url
+
+
 def is_port_open(port_or_url, http_path=None, expect_success=True, protocols=['tcp']):
     port = port_or_url
     host = 'localhost'
@@ -412,12 +426,21 @@ def wait_for_port_open(port, http_path=None, expect_success=True, retries=10, sl
     return retry(check, sleep=sleep_time, retries=retries)
 
 
-def get_free_tcp_port():
-    tcp = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-    tcp.bind(('', 0))
-    addr, port = tcp.getsockname()
-    tcp.close()
-    return port
+def get_free_tcp_port(blacklist=None):
+    blacklist = blacklist or []
+    for i in range(10):
+        tcp = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+        tcp.bind(('', 0))
+        addr, port = tcp.getsockname()
+        tcp.close()
+        if port not in blacklist:
+            return port
+    raise Exception('Unable to determine free TCP port with blacklist %s' % blacklist)
+
+
+def sleep_forever():
+    while True:
+        time.sleep(1)
 
 
 def get_service_protocol():
@@ -434,7 +457,7 @@ def timestamp(time=None, format=TIMESTAMP_FORMAT):
 
 def timestamp_millis(time=None):
     microsecond_time = timestamp(time=time, format=TIMESTAMP_FORMAT_MICROS)
-    # truncating micorseconds to milliseconds. while leaving the Z
+    # truncating microseconds to milliseconds, while leaving the "Z" indicator
     return microsecond_time[:-4] + microsecond_time[-1]
 
 
@@ -636,22 +659,6 @@ def download(url, path, verify_ssl=True):
         s.close()
 
 
-def parse_chunked_data(data):
-    """ Parse the body of an HTTP message transmitted with chunked transfer encoding. """
-    data = (data or '').strip()
-    chunks = []
-    while data:
-        length = re.match(r'^([0-9a-zA-Z]+)\r\n.*', data)
-        if not length:
-            break
-        length = length.group(1).lower()
-        length = int(length, 16)
-        data = data.partition('\r\n')[2]
-        chunks.append(data[:length])
-        data = data[length:].strip()
-    return ''.join(chunks)
-
-
 def first_char_to_lower(s):
     return '%s%s' % (s[0].lower(), s[1:])
 
@@ -699,6 +706,10 @@ def short_uid():
     return str(uuid.uuid4())[0:8]
 
 
+def long_uid():
+    return str(uuid.uuid4())
+
+
 def json_safe(item):
     """ return a copy of the given object (e.g., dict) that is safe for JSON dumping """
     try:
@@ -763,21 +774,38 @@ def cleanup(files=True, env=ENV_DEV, quiet=True):
 
 
 def cleanup_threads_and_processes(quiet=True):
-    for t in TMP_THREADS:
-        t.stop(quiet=quiet)
+    for thread in TMP_THREADS:
+        if thread:
+            try:
+                if hasattr(thread, 'shutdown'):
+                    thread.shutdown()
+                    continue
+                thread.stop(quiet=quiet)
+            except Exception as e:
+                print(e)
     for p in TMP_PROCESSES:
         try:
             p.terminate()
         except Exception as e:
             print(e)
+    # clean up async tasks
+    try:
+        import asyncio
+        for task in asyncio.all_tasks():
+            try:
+                task.cancel()
+            except Exception:
+                pass
+    except Exception:
+        pass
     # clear lists
     clear_list(TMP_THREADS)
     clear_list(TMP_PROCESSES)
 
 
-def clear_list(l):
-    while len(l):
-        del l[0]
+def clear_list(list_obj):
+    while len(list_obj):
+        del list_obj[0]
 
 
 def cleanup_tmp_files():
@@ -904,23 +932,32 @@ def generate_ssl_cert(target_file=None, overwrite=False, random=False, return_co
     def all_exist(*files):
         return all([os.path.exists(f) for f in files])
 
+    def store_cert_key_files(base_filename):
+        key_file_name = '%s.key' % base_filename
+        cert_file_name = '%s.crt' % base_filename
+        # extract key and cert from target_file and store into separate files
+        content = load_file(target_file)
+        key_start = '-----BEGIN PRIVATE KEY-----'
+        key_end = '-----END PRIVATE KEY-----'
+        cert_start = '-----BEGIN CERTIFICATE-----'
+        cert_end = '-----END CERTIFICATE-----'
+        key_content = content[content.index(key_start): content.index(key_end) + len(key_end)]
+        cert_content = content[content.index(cert_start): content.rindex(cert_end) + len(cert_end)]
+        save_file(key_file_name, key_content)
+        save_file(cert_file_name, cert_content)
+        return cert_file_name, key_file_name
+
     if target_file and not overwrite and os.path.exists(target_file):
-        key_file_name = '%s.key' % target_file
-        cert_file_name = '%s.crt' % target_file
+        key_file_name = ''
+        cert_file_name = ''
         try:
-            # extract key and cert from target_file and store into separate files
-            content = load_file(target_file)
-            key_start = '-----BEGIN PRIVATE KEY-----'
-            key_end = '-----END PRIVATE KEY-----'
-            cert_start = '-----BEGIN CERTIFICATE-----'
-            cert_end = '-----END CERTIFICATE-----'
-            key_content = content[content.index(key_start): content.index(key_end) + len(key_end)]
-            cert_content = content[content.index(cert_start): content.index(cert_end) + len(cert_end)]
-            save_file(key_file_name, key_content)
-            save_file(cert_file_name, cert_content)
+            cert_file_name, key_file_name = store_cert_key_files(target_file)
         except Exception as e:
-            LOG.info('Unable to store key/cert files for custom SSL certificate: %s' % e)
-        if all_exist(key_file_name, cert_file_name):
+            # fall back to temporary files if we cannot store/overwrite the files above
+            LOG.info('Error storing key/cert SSL files (falling back to random tmp file names): %s' % e)
+            target_file_tmp = new_tmp_file()
+            cert_file_name, key_file_name = store_cert_key_files(target_file_tmp)
+        if all_exist(cert_file_name, key_file_name):
             return target_file, cert_file_name, key_file_name
     if random and target_file:
         if '.' in target_file:
@@ -1114,7 +1151,7 @@ def clean_cache(file_pattern=CACHE_FILE_PATTERN,
 
 
 def truncate(data, max_length=100):
-    return (data[:max_length] + '...') if len(data) > max_length else data
+    return ('%s...' % data[:max_length]) if len(data or '') > max_length else data
 
 
 def parallelize(func, list, size=None):
